/*
 * $Id: parse_mail.c,v 1.25 1995/05/27 16:56:36 kooper Exp $
 *
 * Copyright (C) 1995, Rob Kooper (kooper@cc.gatech.edu)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <faces.h>
#include "mail.h"

/*
 * local prototypes
 */
#if NeedFunctionPrototypes
char *parseFrom (char *);
void  deleteMail(Mail *);
void  getHeader (char **, char *);
void  copystring(char *, char **, int);
void  copyHeader(char *, char *, char **);
#else
char *parseFrom ();
void  deleteMail();
void  getHeader ();
void  copystring();
void  copyHeader();
#endif


/****************************************************************************
 * This function will parse the mail file and get the From: and the Subject
 * lines from it. It will also try to find a face for the person.
 * It will assume that if a message is not in the list anymore that it has
 * been deleted. This means that a new message to the mailbox is always
 * added at the end of the mailbox
 ****************************************************************************/
#if NeedFunctionPrototypes
int parseMail(void)
#else
int parseMail()
#endif
{
  static time_t  ctime = 0;
  static off_t   size  = 0;
  Mail          *prev, *index;
  FILE          *fp;
  char          *data, *ptr;
  int            newfaces, remfaces;
  struct stat    buf;
  char          *from, *subject, *xface, *MailFile, *mailPtr;

  /*
   * Initialize some variables
   */
  from     = NULL;
  subject  = NULL;
  xface    = NULL;
  ptr      = NULL;
  newfaces = 0;
  remfaces = 0;
  prev     = NULL;
  index    = firstMail;
  data     = XtMalloc(facesRec.MailBuf);

  /*
   * Do a stat on the mailfile, and get modification time.
   */
  if (stat(facesRec.MailFile, &buf) < 0) {
    if (errno != ENOENT) {
      perror("Stat mailfile:");
      exit(-1);
    }

    /*
     * This probably means that the mailfile is gone, so no new mail
     */
    ctime = 0;
    while (index != NULL) {
      prev = index;
      index = prev->next;
      deleteMail(prev);
      remfaces--;
    }
    firstMail = NULL;
    firstFace = NULL;
    return remfaces;
  }

  /*
   * No change in the modification time and the size so no new mail.
   */
  if ((firstMail != NULL) && (buf.st_ctime <= ctime) &&
      (buf.st_size <= size)) {
    if (facesRec.Debug)
      printf("No changes in the mailfile.\n");
    return 0;
  }

  /*
   * Remember the time of change for the next run.
   */
  ctime = buf.st_ctime;
  size  = buf.st_size;

  /*
   * Open the mailfile to read.
   */
  if ((fp = fopen(facesRec.MailFile, "r")) == NULL) {
    perror("Opening mailfile:");
    exit(-1);
  }

  /*
   * Read the mail file into memory.
   */
  MailFile = XtMalloc(size+1);
  if (fread(MailFile, sizeof(char), size, fp) != size)
    fprintf(stderr, "Expected to read %d bytes.\n", (int)size);

  MailFile[size] = 0;
  fclose(fp);

  /*
   * Open the database where the images are stored
   */
  FaceOpenDB();

  /*
   * Check the mailfile
   */
  mailPtr = MailFile;
  if (*mailPtr != 0)
    getHeader(&mailPtr, data);
  while (*mailPtr != 0) {
    /*
     * See if there is a facesRec.MailSeparator.
     */
    while ((strncmp(data, facesRec.MailSep, strlen(facesRec.MailSep)) != 0) &&
	   (*mailPtr != 0))
      getHeader(&mailPtr, data);

    /*
     * If the separator is a from line then get the email address.
     */
    if ((*mailPtr != 0) && (strncmp(facesRec.MailSep, "From", 4) == 0)) {
      copyHeader(data, "From ", &from);
      ptr = from;
      while ((*ptr != 0) && (*ptr != ' '))
	ptr++;
      *ptr = 0;
    }

    /*
     * Read all the headers.
     */
    getHeader(&mailPtr, data);

    while ((strcmp(data, "") != 0) &&
           (strncmp(data, facesRec.MailSep, strlen(facesRec.MailSep)) != 0)) {
      copyHeader(data, "From: ",    &from);
      copyHeader(data, "Subject: ", &subject);
      copyHeader(data, "X-Face: ",  &xface);

      getHeader(&mailPtr, data);
    }

    /*
     * Add the message to the Mail list
     */
    if (from != NULL) {
      ptr  = from;
      from = parseFrom(ptr);
      if (subject == NULL)
        subject = XtNewString("");

      /*
       * Find the first message of that person.
       */
      while ((index != NULL) && (strcmp(index->from, from) != 0) &&
	     (strcmp(index->subject, subject) != 0)) {
	if (prev == NULL)
	  firstMail = index->next;
	else
	  prev->next = index->next;
	if (index->next != NULL)
	  index->next->prev = prev;
	deleteMail(index);
	remfaces--;
	if (prev == NULL)
	  index = firstMail;
	else
	  index = prev->next;
      }

      /*
       * If the message was not found the create a new message.
       */
      if (index == NULL) {
	index          = XtNew(Mail);
	index->from    = XtNewString(from);
	index->subject = XtNewString(subject);
	index->xface   = XtNewString(xface);
	index->next    = NULL;
	index->prev    = prev;

	/*
	 * Count the new messages
	 */
	newfaces++;
      
	/*
	 * Find the face that goes with the address.
	 */
	index->face = FaceSearchDB(from, xface);

	/*
	 * Add the message to the list.
	 */
	if (firstMail == NULL)
	  firstMail = index;
	else
	  prev->next = index;
      }
      prev  = index;
      index = prev->next;
    }

    XtFree(from);    from    = NULL;
    XtFree(ptr);     ptr     = NULL;
    XtFree(subject); subject = NULL;
    XtFree(xface);   xface   = NULL;
  }

  if (prev != NULL)
    prev->next = NULL;

  while(index != NULL) {
    prev = index;
    index = prev->next;
    deleteMail(prev);
    remfaces--;
  }

  FaceCloseDB();
  XtFree(data);
  XtFree(MailFile);

  if (newfaces > 0)
    return newfaces;

  return remfaces;
}

#if NeedFunctionPrototypes
char *parseFrom(char *data)
#else
char *parseFrom(data)
  char *data;
#endif
{
  char *ptr, *result;
  int   quote, back, found, at;

  /*
   * First try:
   *  The address is of the form 'name <address>'
   */
  ptr     = data;
  quote   = 0;
  back    = 0;
  found   = 0;
  at      = 0;
  result  = NULL;

  while (*ptr != 0) {
    if ((*ptr == '"') && !back)
      quote = !quote;
    if ((*ptr == '<') && !back && !quote && (found == 0)) {
      result = ptr+1;
      found = 1;
    }
    if ((*ptr == '>') && !back && !quote && (found == 1)) {
      *ptr = 0;
      found = 2;
    }
    if ((*ptr == '@') && !back && !quote && (found == 1))
      at = 1;
    back = 0;
    if (*ptr == '\\')
      back = 1;
    ptr++;
  }
  if (found == 2) {
    if (at == 0) {
      if (facesRec.Debug)
	printf("LOCAL ADDRESS\n");
      ptr = result;
      result = XtMalloc(strlen(ptr) + strlen(facesRec.Local) + 2);
      strcpy(result, ptr);
      strcat(result, "@");
      strcat(result, facesRec.Local);
      return result;
    }
    return XtNewString(result);
  }

  /*
   * Second try:
   *  The address is of the form 'address (name)'
   */
  ptr     = data;
  quote   = 0;
  back    = 0;
  found   = 0;
  at      = 0;
  result  = ptr;

  while (*ptr != 0) {
    if ((*ptr == '"') && !back)
      quote = !quote;
    if ((*ptr == '@') && !back && !quote && (found == 0))
      at = 1;
    if ((*ptr == '(') && !back && !quote && (found == 0))
      found = 1;
    if ((*ptr == '(') && !back && !quote && (found == 1))
      found = 2;
    back = 0;
    if (*ptr == '\\')
      back = 1;
    ptr++;
  }
  if (found == 2) {
    ptr = result;
    while ((*ptr != ' ') && (*ptr != 0))
      ptr++;
    *ptr = 0;
    if (at == 0) {
      if (facesRec.Debug)
	printf("LOCAL ADDRESS\n");
      ptr = result;
      result = XtMalloc(strlen(ptr) + strlen(facesRec.Local) + 2);
      strcpy(result, ptr);
      strcat(result, "@");
      strcat(result, facesRec.Local);
      return result;
    }
    return XtNewString(result);
  }

  /*
   * Third try:
   *   There is no name just an address.
   */
  ptr     = data;
  quote   = 0;
  back    = 0;
  found   = 0;
  at      = 0;
  result  = ptr;

  while(*ptr != 0) {
    if ((*ptr == '"') && !back)
      quote = !quote;
    if ((*ptr == '@') && !back && !quote)
      at = 1;
    if (((*ptr == ' ') || (*ptr == '\n')) && !back && !quote)
      *ptr = 0;
    back = 0;
    if (*ptr == '\\')
      back = 1;
    ptr++;
  }
  if (at == 0) {
    if (facesRec.Debug)
      printf("LOCAL ADDRESS\n");
    ptr = result;
    result = XtMalloc(strlen(ptr) + strlen(facesRec.Local) + 2);
    strcpy(result, ptr);
    strcat(result, "@");
    strcat(result, facesRec.Local);
    return result;
  }
  return XtNewString(result);
}

#if NeedFunctionPrototypes
void deleteMail(Mail *delete)
#else
void deleteMail(delete)
  Mail *delete;
#endif
{
  Face *face;

  if (facesRec.Debug)
    printf("Removing mail from :%s\n", delete->from);

  face = delete->face;
  XtFree(delete->from);
  XtFree(delete->subject);
  XtFree((char *)delete);

  FaceDelete(face);
}

#if NeedFunctionPrototypes
void getHeader(char **mailPtr, char *data)
#else
void getHeader(mailPtr, data)
  char **mailPtr, *data;
#endif
{
  int   length;

  strcpy(data, "");
  length = 0;

  /*
   * Return if there is an end of file
   */
  if (*mailPtr == 0)
    return;

  /*
   * Read the next header.
   */
  copystring(data, mailPtr, facesRec.MailBuf);
  length = strlen(data);

  /*
   * See if it is an empty '\n', if so return '' indicating end of headers.
   */
  if (strcmp(data, "") == 0)
    return;

  /*
   * Keep reading if the first character of the next line is either a
   * ' ' or a tab.
   */
  while (((**mailPtr == ' ') || (**mailPtr == '\t')) &&
	 (length < facesRec.MailBuf)) {
    copystring((data+length), mailPtr, facesRec.MailBuf-length);
    length = strlen(data);
  }
}

#if NeedFunctionPrototypes
void copystring(char *dst, char **src, int n)
#else
void copystring(dst, src, n)
  char *dst, **src;
  int   n;
#endif
{
  while ((**src != 0) && (**src != '\n') && (n > 0)) {
    *dst++ = **src;
    (*src)++;
    n--;
  }

  *dst = 0;
  if (**src == '\n')
    (*src)++;
}

#if NeedFunctionPrototypes
void  copyHeader(char *data, char *search, char **result)
#else
void  copyHeader(data, search, result)
  char *data, *search, **result;
#endif
{
  int length;

  /*
   * See if the data starts with search string.
   */
  length = strlen(search);
  if (strncmp(data, search, length) != 0)
    return;

  /*
   * Strip the search string of the data and put it in result.
   */
  *result = XtMalloc(strlen(data) - length);
  strcpy(*result, (data+length));
}
