/*
 * $Id: find_face.c,v 1.29 1995/05/30 20:52:17 kooper Exp $
 *
 * Copyright (C) 1995, Rob Kooper (kooper@cc.gatech.edu)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include "facesP.h"

/*
 * local structures
 */
typedef struct Contents {
  char            *Major,
                  *Minor,
                  *Data;
  int              Length;
  struct Contents *next;
} Contents;

/*
 * local prototypes
 */
#if NeedFunctionPrototypes
void      face_searchDB       (Face **, datum, char *);
void      face_breadthSearchDB(Face **, char *, char *);
void      face_depthSearchDB  (Face **, char *, char *);
void      face_knownSearchDB  (Face **, char *, char *);
void      face_lowercase      (char *);
void      face_searchLink     (Face **, char *);
void      face_foundIt        (Face **, datum, char *);
void      face_searchFaces    (Face **, char *);
Boolean   face_nextQuery      (char **, char *);
Contents *face_extractContents(datum);
#else
void      face_searchDB       ();
void      face_breadthSearchDB();
void      face_depthSearchDB  ();
void      face_knownSearchDB  ();
void      face_lowercase      ();
void      face_searchLink     ();
void      face_foundIt        ();
void      face_searchFaces    ();
Boolean   face_nextQuery      ();
Contents *face_extractContents();
#endif

/*************************************************************************
 * This function will call the right subfunction, depending on the
 *  facesOptions.SearchFunction
 *************************************************************************/
#if NeedFunctionPrototypes
Face *FaceSearchDB(char *address, char *xface)
#else
Face *FaceSearchDB(address, xface)
  char *address, *xface;
#endif
{
  Face *face;

  face = NULL;

  switch(facesOptions.SearchFunction) {
  case searchDepth   :
    face_depthSearchDB(&face, address, xface);
    break;
  case searchBreadth :
    face_breadthSearchDB(&face, address, xface);
    break;
  case searchKnown   :    
    face_knownSearchDB(&face, address, xface);
    break;
  default            :
    WARNING(("Don't know about function %d", facesOptions.SearchFunction));
    break;
  }

  /*
   * Add the face to the list and return.
   */
  face_addFace(face);
  return face;
}

/*************************************************************************
 * This function will search all the databases first before moving on
 * to the next query.
 *************************************************************************/
#if NeedFunctionPrototypes
void face_breadthSearchDB(Face **facePtr, char *address, char *xface)
#else
void face_breadthSearchDB(facePtr, address, xface)
  Face **facePtr;
  char  *address, *xface;
#endif
{
  Face      *face;
  char      *find;
  char      *loweraddress;
  Boolean    openDatabase, loadXface;
  FacesDB   *index;
  datum      key, content;

  face = *facePtr;

  /*
   * See if the database is open
   */
  if (FacesPrivate.databaseFlags != 2) {
    FaceOpenDB();
    openDatabase = True;
  } else
    openDatabase = False;

  /*
   * Turn the address into lower case
   */
  loweraddress = XtNewString(address);
  face_lowercase(loweraddress);

  /*
   * See if the xface should be loaded.
   */
#ifdef XFACE
  if (facesOptions.XfaceFirst && (xface != NULL) && (face == NULL)) {
    face_searchFaces(&face, loweraddress);
    face_loadXFace(&face, xface);
    if (face != NULL) {
      if (facesOptions.Debug)
	printf("Created face from xface data.\n");
      face->name = XtNewString(loweraddress);
    }
  }
#endif

  /*
   * Go find the face and load it.
   */
  find      = XtNewString(loweraddress);
  loadXface = False;

  while ((strcmp(find, "") != 0) &&
	 ((face == NULL) || (face->icon == None))) {
    /*
     * See if the xface should be loaded.
     */
#ifdef XFACE
    if (loadXface && !facesOptions.XfaceFirst && (xface != NULL)) {
      face_searchFaces(&face, loweraddress);
      face_loadXFace(&face, xface);
      if ((face != NULL) && (face->name == NULL)) {
	if (facesOptions.Debug)
	  printf("Created face from xface data.\n");
	face->name = XtNewString(loweraddress);
      }
    }
#endif

    if (facesOptions.Debug)
      printf("Searching for: %s.\n", find);

    key.dptr      = find;
    key.dsize     = strlen(find);
    content.dptr  = NULL;
    content.dsize = 0;

    /*
     * Search all the databases
     */
    index = FacesPrivate.firstDB;
    while ((index != NULL) && ((face == NULL) || (face->icon == None))) {
      if (facesOptions.Debug && (index != NULL))
	printf("Using database %s.\n", index->path);

      if (index->dbf == NULL) {
	openDatabase = True;
	face_openDB(index);
      } else
	openDatabase = False;

      if (index->dbf != NULL) {
	content = gdbm_fetch(index->dbf, key);
	if (content.dptr != NULL) {
	  /*
	   * Something was found, display it.
	   */
	  face_searchDB(&face, content, find);
	  if ((face != NULL) && (face->name == NULL))
	    face->name = XtNewString(find);
	}

	if (openDatabase)
	  face_closeDB(index);
      }
      index = index->next;
    }

    /*
     * Formulate the next query.
     */
    loadXface = face_nextQuery(&find, loweraddress);
  }

  /*
   * If the face was not found return the unknown face
   */
  if ((face == NULL) || (face->icon == None)) {
    if (facesOptions.Debug)
      printf("%s not found return unknownface.\n", address);
    face_searchFaces(&face, "unknown");
    if ((face == NULL) || (face->icon == None))
      face_loadUnknown(&face);
  }

  XtFree(find);
  XtFree(loweraddress);

  if (openDatabase)
    FaceCloseDB();

  *facePtr = face;
}

/*************************************************************************
 * This function will search all queries in a database before moving
 * to the next database.
 *************************************************************************/
#if NeedFunctionPrototypes
void face_depthSearchDB(Face **facePtr, char *address, char *xface)
#else
void face_depthSearchDB(facePtr, address, xface)
  Face **facePtr;
  char  *address, *xface;
#endif
{
  Face      *face;
  char      *find;
  char      *loweraddress;
  Boolean    openDatabase, loadXface;
  FacesDB   *index;
  datum      key, content;

  face = *facePtr;

  /*
   * See if the database is initialized
   */
  if (FacesPrivate.databaseFlags == 0)
    FaceInitDB();

  /*
   * Turn the address into lower case
   */
  loweraddress = XtNewString(address);
  face_lowercase(loweraddress);

  /*
   * See if the xface should be loaded.
   */
#ifdef XFACE
  if (facesOptions.XfaceFirst && (xface != NULL) && (face == NULL)) {
    face_searchFaces(&face, loweraddress);
    face_loadXFace(&face, xface);
    if (face != NULL) {
      if (facesOptions.Debug)
	printf("Created face from xface data.\n");
      face->name = XtNewString(loweraddress);
    }
  }
#endif

  /*
   * Go find the face and load it.
   */
  index = FacesPrivate.firstDB;

  while (((face == NULL) || (face->icon == None)) && (index != NULL)) {
    find      = XtNewString(loweraddress);
    loadXface = False;

    if (facesOptions.Debug && (index != NULL))
      printf("Using database %s.\n", index->path);

    if (index->dbf == NULL) {
      openDatabase = True;
      face_openDB(index);
    } else
      openDatabase = False;

    while ((index->dbf != NULL) && ((face == NULL) || (face->icon == None)) &&
	   (strcmp(find, "") != 0)) {
      /*
       * See if the xface should be loaded.
       */
#ifdef XFACE
      if (loadXface && !facesOptions.XfaceFirst && (xface != NULL)) {
	face_searchFaces(&face, loweraddress);
	face_loadXFace(&face, xface);
	if ((face != NULL) && (face->name == NULL)) {
	  if (facesOptions.Debug)
	    printf("Created face from xface data.\n");
	  face->name = XtNewString(loweraddress);
	}
      }
#endif

      if (facesOptions.Debug)
	printf("New query is %s.\n", find);

      key.dptr      = find;
      key.dsize     = strlen(find);
      content.dptr  = NULL;
      content.dsize = 0;
      
      /*
       * Search the database for the particular face.
       */
      content = gdbm_fetch(index->dbf, key);
      if (content.dptr != NULL) {
	/*
	 * Something was found, display it.
	 */
	face_searchDB(&face, content, find);
	if ((face != NULL) && (face->name == NULL))
	  face->name = XtNewString(find);
      }

      /*
       * Formulate the next query.
       */
      loadXface = face_nextQuery(&find, loweraddress);
    }

    /*
     * Close the database if we opened it.
     */
    if (openDatabase && (index->dbf != NULL))
      face_closeDB(index);

    index = index->next;
  }

  /*
   * If the face was not found return the unknown face
   */
  if (face == NULL) {
    if (facesOptions.Debug)
      printf("%s not found return unknownface.\n", address);
    face_searchFaces(&face, "unknown");
    if (face == NULL)
      face_loadUnknown(&face);
  }

  XtFree(find);
  XtFree(loweraddress);

  *facePtr = face;
}

/*************************************************************************
 * This function will search all known queries in a database before
 * moving to the next database. Then it will start over again but with
 * the unknown queries.
 *************************************************************************/
#if NeedFunctionPrototypes
void face_knownSearchDB(Face **facePtr, char *address, char *xface)
#else
void face_knownSearchDB(facePtr, address, xface)
  Face **facePtr;
  char  *address, *xface;
#endif
{
  Face      *face;
  char      *find;
  char      *loweraddress;
  Boolean    openDatabase, loadXface, unknown;
  FacesDB   *index;
  datum      key, content;

  face = *facePtr;

  /*
   * See if the database is initialized
   */
  if (FacesPrivate.databaseFlags == 0)
    FaceInitDB();

  /*
   * Turn the address into lower case
   */
  loweraddress = XtNewString(address);
  face_lowercase(loweraddress);

  /*
   * See if the xface should be loaded.
   */
#ifdef XFACE
  if (facesOptions.XfaceFirst && (xface != NULL) && (face == NULL)) {
    face_searchFaces(&face, loweraddress);
    face_loadXFace(&face, xface);
    if (face != NULL) {
      if (facesOptions.Debug)
	printf("Created face from xface data.\n");
      face->name = XtNewString(loweraddress);
    }
  }
#endif

  /*
   * Go find the face and load it.
   */
  unknown   = False;
  loadXface = False;
  find      = NULL;
  while (((face == NULL) || (face->icon == None)) && !unknown) {
    /*
     * Set the flag to signal this is the second time. And create a new
     * loweraddress (unknown@bla.bla)
     */
    if (loadXface) {
      unknown      = True;
      loweraddress = XtNewString(find);
    }

    /*
     * Search all the databases
     */
    index = FacesPrivate.firstDB;
    while (((face == NULL) || (face->icon == None)) && (index != NULL)) {
      loadXface = False;
      find      = XtNewString(loweraddress);
      
      if (facesOptions.Debug && (index != NULL))
	printf("Using database %s.\n", index->path);

      if (index->dbf == NULL) {
	openDatabase = True;
	face_openDB(index);
      } else
	openDatabase = False;

      while ((index->dbf != NULL) && ((face == NULL) || (face->icon == None))
	     && (strcmp(find, "") != 0) && !loadXface) {

	if (facesOptions.Debug)
	  printf("Searching for: %s.\n", find);

	key.dptr      = find;
	key.dsize     = strlen(find);
	content.dptr  = NULL;
	content.dsize = 0;
      
	/*
	 * Search the database for the particular face.
	 */
	content = gdbm_fetch(index->dbf, key);
	if (content.dptr != NULL) {
	  /*
	   * Something was found, display it.
	   */
	  face_searchDB(&face, content, find);
	  if ((face != NULL) && (face->name == NULL))
	    face->name = XtNewString(find);
	}
	
	/*
	 * Formulate the next query.
	 */
	loadXface = face_nextQuery(&find, loweraddress);
      }

      /*
       * Close the database if we opened it.
       */
      if (openDatabase && (index->dbf != NULL))
	face_closeDB(index);

      index = index->next;
    }
    /*
     * See if the xface should be loaded.
     */
#ifdef XFACE
    if (loadXface && !facesOptions.XfaceFirst && (xface != NULL)) {
      face_searchFaces(&face, loweraddress);
      face_loadXFace(&face, xface);
      if ((face != NULL) && (face->name == NULL)) {
	if (facesOptions.Debug)
	  printf("Created face from xface data.\n");
	face->name = XtNewString(loweraddress);
      }
    }
#endif
  }

  /*
   * If the face was not found return the unknown face
   */
  if (face == NULL) {
    if (facesOptions.Debug)
      printf("%s not found return unknownface.\n", address);
    face_searchFaces(&face, "unknown");
    if (face == NULL)
      face_loadUnknown(&face);
  }

  XtFree(find);
  XtFree(loweraddress);

  *facePtr = face;
}

/*************************************************************************
 * This function will turn the address to lowercase, but will leave the
 * MISC intact.
 *************************************************************************/
#if NeedFunctionPrototypes
void face_lowercase(char *address)
#else
void face_lowercase(address)
  char *address;
#endif
{
  char *index;

  if (address == NULL)
    return;

  /*
   * Turn the address into lower case
   */
  index = address;
  while ((*index != 0) && (strcmp(index, "MISC") != 0)) {
    if (isupper(*index))
      *index = tolower(*index);
    index++;
  }
}

/*************************************************************************
 * This function will check to see if the entry is a link or a face, and
 * act accordingly.
 *************************************************************************/
#if NeedFunctionPrototypes
void face_searchDB(Face **facePtr, datum content, char *name)
#else
void face_searchDB(facePtr, content, name)
  Face  **facePtr;
  datum   content;
  char   *name;
#endif
{
  if (strncmp(content.dptr, "link:", 5) == 0) {
    /*
     * It's a link to another face in the database.
     */
    if (facesOptions.Debug)
      printf("Found a link to %s\n", (content.dptr)+5);
    
    face_searchLink(facePtr, (content.dptr)+5);
  } else {
    /*
     * It's a normal face.
     */
    face_foundIt(facePtr, content, name);
  }
}

/*************************************************************************
 * This will search all the databases and see if there exists a face.
 *************************************************************************/
#if NeedFunctionPrototypes
void face_searchLink(Face **facePtr, char *address)
#else
void face_searchLink(facePtr, address)
  Face **facePtr;
  char  *address;
#endif
{
  Boolean  wasOpen;
  FacesDB *index;
  datum    key, content;

  key.dptr      = address;
  key.dsize     = strlen(address);
  content.dptr  = NULL;
  content.dsize = 0;  
  index         = FacesPrivate.firstDB;

  while (index != NULL) {
    if (index->dbf == NULL) {
      wasOpen = False;
      face_openDB(index);
    } else
      wasOpen = True;

    if (index->dbf != NULL) {
      content = gdbm_fetch(index->dbf, key);
      /*
       * Something was found, display it.
       */
      if (content.dptr != NULL)
	face_searchDB(facePtr, content, address);

      if (!wasOpen)
	face_closeDB(index);
    }

    index = index->next;
  }
}

/*************************************************************************
 * This will create a new query along the following lines:
 *  kooper@blabla.cc.gatech.edu
 *  kooper@cc.gatech.edu
 *  kooper@gatech.edu
 *  kooper@edu
 *  kooper@MISC
 *  unknonw@blabla.cc.gatech.edu
 *  unknonw@cc.gatech.edu
 *  unknonw@gatech.edu
 *  unknonw@edu
 *  unknonw@MISC
 *************************************************************************/
#if NeedFunctionPrototypes
Boolean face_nextQuery(char **from, char *orig)
#else
Boolean face_nextQuery(from, orig)
  char **from, *orig;
#endif
{
  char    *name, *net, *src, *dest;
  
  /*
   * Return if nothing needs to be done.
   */
  if (strcmp(*from, "") == 0)
    return False;

  /*
   * Find the '@' and replace it with 0.
   */
  dest    = strchr(*from, '@');
  *dest++ = 0;

  /*
   * Copy the name.
   */
  name = XtNewString(*from);

  /*
   * Find the first . since the @
   */
  src  = strchr(dest, '.');
  if (src == NULL) {
    if (strcmp(dest, "MISC") == 0) {
      if (strcmp(name, "unknown") == 0) {
	/*
	 * No more queries, return ""
	 */
	*from = XtRealloc(*from, 1);
	**from = 0;
	XtFree(name);
	return False;
      } else {
	/*
	 * Start the unknown queries
	 */
	XtFree(name);
	name    = XtNewString(orig);
	net     = strchr(name, '@');
	*from = XtRealloc(*from, strlen(net) + 8);
	strcpy(*from, "unknown");
	strcat(*from, net);
	XtFree(name);
	return True;
      }
    } else {
      *from = XtRealloc(*from, strlen(name) + 6);
      strcpy(*from, name);
      strcat(*from, "@MISC");
      XtFree(name);
      return False;
    }
  }

  /*
   * Make a copy of the rest of the address.
   */
  *src++ = 0;
  net = XtNewString(src);

  /*
   * Put it all back together.
   */
  *from = XtRealloc(*from, strlen(name) + strlen(net) + 1);
  strcpy(*from, name);
  strcat(*from, "@");
  strcat(*from, net);

  XtFree(name);
  XtFree(net);

  return False;
}

/*************************************************************************
 * If a face was found this function is called and looks at the entry.
 * It will try to load the images in order and at the end display the
 * rest of the data.
 *************************************************************************/
#if NeedFunctionPrototypes
void face_foundIt(Face **facePtr, datum info, char *name)
#else
void face_foundIt(facePtr, info, name)
  Face  **facePtr;
  datum   info;
  char   *name;
#endif
{
  char     *startType, *firsttype, *nexttype;
  Contents *contents, *index;

  /*
   * See if the face is already loaded.
   */
  face_searchFaces(facePtr, name);
  
  /*
   * What type first.
   */
  startType = XtNewString(facesOptions.ImageOrder);
  firsttype = startType;

  /*
   * Extract all the contents of the entry.
   */
  contents = face_extractContents(info);

  while (firsttype != NULL) {
    /*
     * Determine the next type.
     */
    nexttype  = strchr(firsttype, ':');
    if (nexttype != NULL) 
      *nexttype++ = 0;

    index = contents;
    while (index != NULL) {
      if (facesOptions.Debug)
	printf("It is an %s of type %s with length %d\n",
	       index->Major, index->Minor, index->Length);
      
      /*
       * Use the data, using the right program (.mailcap)
       */
      if (strcmp(firsttype, index->Minor) == 0)
	face_useMailcap(facePtr, index->Data, index->Length,
			index->Major, index->Minor, True);
      else if (strcmp(firsttype, "*") == 0)
	face_useMailcap(facePtr, index->Data, index->Length,
			index->Major, index->Minor, False);

      index = index->next;

    }

    firsttype = nexttype;
  }

  /*
   * Remove all the temp vars
   */
  while (contents != NULL) {
    index = contents->next;
    XtFree((char *) contents);
    contents = index;
  }

  XtFree(startType);
}

/*************************************************************************
 * This function checks to see if a face is already loaded. This way a
 * face is loaded only once.
 *************************************************************************/
#if NeedFunctionPrototypes
void face_searchFaces(Face **facePtr, char *address)
#else
void face_searchFaces(facePtr, address)
  Face **facePtr;
  char  *address;
#endif
{
  Face *index, *face;

  index = firstFace;
  face  = *facePtr;

  /*
   * See if there is already a face with this address, then just
   * return a pointer to the face already loaded.
   */
  while (index != NULL) {

    if (strcmp(index->name, address) == 0) {
      if (facesOptions.Debug)
	printf("Reusing data for %s.\n", address);
      index->counter++;
      if (face == NULL) {
	*facePtr = index;
	return;
      }
      face->iconPtr    = index;
      face->icon_width = index->icon_width;
      face->icon_width = index->icon_width;
      face->icon       = index->icon;
      return;
    }

    index = index->next;
  }

  return;
}

/*************************************************************************
 * Get all the contents in an entry into some easy structure.
 *************************************************************************/
#if NeedFunctionPrototypes
Contents *face_extractContents(datum content)
#else
Contents *face_extractContents(content)
  datum content;
#endif
{
  datum     orig;
  Contents *result, *next, *prev;
  char     *ptr;

  result     = NULL;
  prev       = NULL;
  orig.dsize = content.dsize;
  orig.dptr  = content.dptr;
  
  while (orig.dsize > 0) {
    next         = XtNew(Contents);
    next->Major  = orig.dptr;
    next->Minor  = strchr(next->Major, '/'); *(next->Minor)++ = 0;
    ptr          = strchr(next->Minor, ':'); *ptr++           = 0;
    next->Data   = strchr(ptr, ':');         *(next->Data)++  = 0;
    next->Length = atol(ptr);
    next->next   = NULL;

    orig.dptr   = next->Data + next->Length;
    if (next->Length == 0)
      orig.dptr += strlen(next->Data);
    orig.dsize -= (orig.dptr - next->Major);

    if (result == NULL)
      result = next;
    else
      prev->next = next;
    
    prev = next;
  }

  return result;
}

