/*
 * $Id: facesdb_xaw.c,v 1.9 1995/03/19 17:16:14 kooper Exp $
 *
 * Copyright (C) 1995, Rob Kooper (kooper@cc.gatech.edu)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "facesdb.h"
#include "facesdb_res.h"
#include "../patchlevel.h"

#include <stdlib.h>
#include <sys/stat.h>

#include <X11/Shell.h>
#include <X11/IntrinsicP.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Label.h>

static String fallbackResources[] = {
#include "FacesDB.ad.h"
  NULL
};

Display       *display;
Window         window;
GC             gc;
XtAppContext   app;

/*
 * Input widgets
 */
static Widget  dialog, databaseInput, addressInput, imageInput, majorInput,
               minorInput, linktoInput, toplevel;
Boolean        dialogPop, useX, noTxt;
char          *progname;
/*
 * Local Prototypes.
 */
#if NeedFunctionPrototypes
int   initDisplay    (int *, char **);
void  pushAdd        (Widget, XtPointer, XtPointer);
void  pushLink       (Widget, XtPointer, XtPointer);
void  pushList       (Widget, XtPointer, XtPointer);
void  pushExtract    (Widget, XtPointer, XtPointer);
void  pushDelete     (Widget, XtPointer, XtPointer);
void  pushQuit       (Widget, XtPointer, XtPointer);
void  dialogString   (char *, char *, char *);
void  dialogButton   (Widget, XtPointer, XtPointer);
char *getWidgetString(Widget);
void  setWidgetString(Widget, char *);
void  usage          (void);
#else
int   initDisplay    ();
void  pushAdd        ();
void  pushLink       ();
void  pushList       ();
void  pushExtract    ();
void  pushDelete     ();
void  pushQuit       ();
void  dialogString   ();
void  dialogButton   ();
char *getWidgetString();
void  setWidgetString();
void  usage          ();
#endif

/*************************************************************************
 *
 *************************************************************************/
#if NeedFunctionPrototypes
void main(int argc, char **argv)
#else
void main(argc, argv)
  int    argc;
  char **argv;
#endif
{
  int function;

  /*
   * Initialize display (get the command line arguments)
   */
  useX     = False;
  function = initDisplay(&argc, argv);
  progname = argv[0];

  /*
   * No more commandline options should be left, otherwise print help.
   */
  if (argc > 1)
    usage();  

  /*
   * See what function the user want and call the right routine.
   */
  switch(function) {
  case FuncNone:
    useX = True;
    XtRealizeWidget(toplevel);
    XtAppMainLoop(app);
    break;
    
  case FuncList:
    pushList(NULL, NULL, NULL);
    break;
    
  case FuncAdd:
    pushAdd(NULL, "Import", NULL);
    break;
    
  case FuncDelete:
    pushDelete(NULL, "Delete", NULL);
    break;
    
  case FuncExtract:
    pushExtract(NULL, NULL, NULL);
    break;

  case FuncLink:
    pushLink(NULL, "Link", NULL);
    break;

  case FuncConvert:
    break;

  case FuncVersion:
    printf("%s Version %s\n", progname, FACES_VERSION);
    break;

  default:
    fprintf(stderr, "This should not happen\n");
    break;
  }
}

/*************************************************************************
 *
 *************************************************************************/
#if NeedFunctionPrototypes
int initDisplay(int *argc, char **argv)
#else
int initDisplay(argc, argv)
  int   *argc;
  char **argv;
#endif
{
  Widget    box, addButton, linkButton, deleteButton, listButton, quitButton,
            extractButton;
  Resources FacesdbRec;

  /* Create an application */
  toplevel = XtAppInitialize(&app, "FacesDB", options, XtNumber(options),
                             argc, argv, fallbackResources, NULL, 0);

  /* Get all the resources */
  XtGetApplicationResources(toplevel, &FacesdbRec,
                            resources, XtNumber(resources),
                            NULL, 0);

  /* create box widget */
  box = XtCreateManagedWidget("box", boxWidgetClass, toplevel, NULL, 0);

  /* Database input field. */
  XtCreateManagedWidget("databaseLabel", labelWidgetClass, box, NULL, 0);
  databaseInput = XtCreateManagedWidget("databaseInput", asciiTextWidgetClass,
					box, NULL, 0);
  setWidgetString(databaseInput, FacesdbRec.Database);

  /* Email input field. */
  XtCreateManagedWidget("addressLabel", labelWidgetClass, box, NULL, 0);
  addressInput = XtCreateManagedWidget("addressInput", asciiTextWidgetClass,
				       box, NULL, 0);
  setWidgetString(addressInput, FacesdbRec.Address);

  /* Image input field. */
  XtCreateManagedWidget("imageLabel", labelWidgetClass, box, NULL, 0);
  imageInput = XtCreateManagedWidget("imageInput", asciiTextWidgetClass,
				     box, NULL, 0);
  setWidgetString(imageInput, FacesdbRec.Image);

  /* Major input field. */
  XtCreateManagedWidget("majorLabel", labelWidgetClass, box, NULL, 0);
  majorInput = XtCreateManagedWidget("majorInput", asciiTextWidgetClass,
				     box, NULL, 0);
  setWidgetString(majorInput, FacesdbRec.Major);

  /* Minor input field. */
  XtCreateManagedWidget("minorLabel", labelWidgetClass, box, NULL, 0);
  minorInput = XtCreateManagedWidget("minorInput", asciiTextWidgetClass,
				     box, NULL, 0);
  setWidgetString(minorInput, FacesdbRec.Minor);

  /* LinkTo input field. */
  XtCreateManagedWidget("linktoLabel", labelWidgetClass, box, NULL, 0);
  linktoInput = XtCreateManagedWidget("linktoInput", asciiTextWidgetClass,
				      box, NULL, 0);
  setWidgetString(linktoInput, FacesdbRec.LinkTo);

  /* add buttons */
  addButton = XtCreateManagedWidget("addButton", commandWidgetClass,
				    box, NULL, 0);
  XtAddCallback(addButton, XtNcallback, pushAdd, NULL);
  
  linkButton = XtCreateManagedWidget("linkButton", commandWidgetClass,
				     box, NULL, 0);
  XtAddCallback(linkButton, XtNcallback, pushLink, NULL);
  
  listButton = XtCreateManagedWidget("listButton", commandWidgetClass,
				     box, NULL, 0);
  XtAddCallback(listButton, XtNcallback, pushList, NULL);
  
  extractButton = XtCreateManagedWidget("extractButton", commandWidgetClass,
				     box, NULL, 0);
  XtAddCallback(extractButton, XtNcallback, pushExtract, NULL);
  
  deleteButton = XtCreateManagedWidget("deleteButton", commandWidgetClass,
				       box, NULL, 0);
  XtAddCallback(deleteButton, XtNcallback, pushDelete, NULL);
  
  quitButton = XtCreateManagedWidget("quitButton", commandWidgetClass,
				     box, NULL, 0);
  XtAddCallback(quitButton, XtNcallback, pushQuit, NULL);

  /* See if there should be any output. */
  noTxt = FacesdbRec.NoTxt;

  /* What function has the program to perform. */
  if (FacesdbRec.Convert != NULL) {
    convertDatabase(FacesdbRec.Convert);
    return FuncConvert;
  }
  return FacesdbRec.Function;
}

/*************************************************************************
 *
 *************************************************************************/
#if NeedFunctionPrototypes
void pushAdd(Widget button, XtPointer userData, XtPointer call_data)
#else
void pushAdd(button, userData, call_data)
  Widget    button;
  XtPointer userData, call_data;
#endif
{
  char        *dbase, *email, *image, *Major, *Minor;
  struct stat  buf;

  /* get the character string in the text field */
  if ((dbase = getWidgetString(databaseInput)) == NULL || *dbase == '\0') {
    dialogString("Need a database name.", "OK", NULL);
    return;
  }

  if ((email = getWidgetString(addressInput)) == NULL || *email == '\0') {
    dialogString("Need an email address.", "OK", NULL);
    return;
  }
  lowerAddress(email);

  if ((image = getWidgetString(imageInput)) == NULL || *image == '\0') {
    dialogString("Need an image filename.", "OK", NULL);
    return;
  }
  if (stat(image, &buf) < 0) {
    dialogString("Could not find image.", "OK", NULL);
    return;
  }

  if ((Major = getWidgetString(majorInput)) == NULL || *Major == '\0') {
    dialogString("Need a major type.", "OK", NULL);
    return;
  }

  if ((Minor = getWidgetString(minorInput)) == NULL || *Minor == '\0') {
    dialogString("Need a minor type.", "OK", NULL);
    return;
  }

  if (userData == NULL)
    dialogString("Import the file in the database?", "Import", "Filename");
  else {
    if (strcmp(userData, "Import") == 0)
      addEntry(dbase, email, image, Major, Minor, buf.st_size);
    else
      addEntry(dbase, email, image, Major, Minor, 0);
  }
}

/*************************************************************************
 *
 *************************************************************************/
#if NeedFunctionPrototypes
void pushLink(Widget button, XtPointer userData, XtPointer call_data)
#else
void pushLink(button, userData, call_data)
  Widget    button;
  XtPointer userData, call_data;
#endif
{
  char        *dbase, *email, *link;

  /* get the character string in the text field */
  if ((dbase = getWidgetString(databaseInput)) == NULL || *dbase == '\0') {
    dialogString("Need a database name.", "OK", NULL);
    return;
  }

  if ((email = getWidgetString(addressInput)) == NULL || *email == '\0') {
    dialogString("Need an email address.", "OK", NULL);
    return;
  }
  lowerAddress(email);

  if ((link = getWidgetString(linktoInput)) == NULL || *link == '\0') {
    dialogString("Need an address to link to.", "OK", NULL);
    return;
  }
  lowerAddress(link);

  if (!checkAddress(dbase, email) || (userData != NULL))
    addLink(dbase, email, link);
  else
    dialogString("Replace existing data with Link?", "Link", "No");
}

/*************************************************************************
 *
 *************************************************************************/
#if NeedFunctionPrototypes
void pushList(Widget button, XtPointer userData, XtPointer call_data)
#else
void pushList(button, userData, call_data)
  Widget    button;
  XtPointer userData, call_data;
#endif
{
  char        *dbase, *email, *ptr;
  struct stat  buf;

  /* get the character string in the text field */
  if ((dbase = getWidgetString(databaseInput)) == NULL || *dbase == '\0') {
    dialogString("Need a database name.", "OK", NULL);
    return;
  }
  if (stat(dbase, &buf) < 0) {
    dialogString("Could not find database.", "OK", NULL);
    return;
  }

  ptr = getWidgetString(addressInput);
  if (*ptr == '\0')
    email = NULL;
  else {
    email = ptr;
    lowerAddress(email);
  }

  /*
   * Check if this is in the database.
   */
  if ((email != NULL) && !checkAddress(dbase, email)) {
    dialogString("This name is not in the database.", "Ok", NULL);
    return;
  }

  listEntry(dbase, email);
}

/*************************************************************************
 *
 *************************************************************************/
#if NeedFunctionPrototypes
void pushExtract(Widget button, XtPointer userData, XtPointer call_data)
#else
void pushExtract(button, userData, call_data)
  Widget    button;
  XtPointer userData, call_data;
#endif
{
  char        *dbase, *email, *image, *Major, *Minor, *ptr;
  struct stat  buf;

  /* get the character string in the text field */
  if ((dbase = getWidgetString(databaseInput)) == NULL || *dbase == '\0') {
    dialogString("Need a database name.", "OK", NULL);
    return;
  }
  if (stat(dbase, &buf) < 0) {
    dialogString("Could not find database.", "OK", NULL);
    return;
  }

  if ((email = getWidgetString(addressInput)) == NULL || *email == '\0') {
    dialogString("Need an email address.", "OK", NULL);
    return;
  }
  lowerAddress(email);
  
  ptr = getWidgetString(imageInput);
  if (*ptr == '\0')
    image = NULL;
  else
    image = ptr;

  ptr = getWidgetString(majorInput);
  if (*ptr == '\0')
    Major = NULL;
  else
    Major = ptr;

  ptr = getWidgetString(minorInput);
  if (*ptr == '\0')
    Minor = NULL;
  else
    Minor = ptr;

  /*
   * Check if this is in the database.
   */
  if (!checkAddress(dbase, email)) {
    dialogString("This name is not in the database.", "Ok", NULL);
    return;
  }

  extractEntry(dbase, email, image, Major, Minor);
}

/*************************************************************************
 *
 *************************************************************************/
#if NeedFunctionPrototypes
void pushDelete(Widget button, XtPointer userData, XtPointer call_data)
#else
void pushDelete(button, userData, call_data)
  Widget    button;
  XtPointer userData, call_data;
#endif
{
  char        *dbase, *email, *Major, *Minor, *ptr;
  struct stat  buf;

  /* get the character string in the text field */
  if ((dbase = getWidgetString(databaseInput)) == NULL || *dbase == '\0') {
    dialogString("Need a database name.", "OK", NULL);
    return;
  }
  if (stat(dbase, &buf) < 0) {
    dialogString("Could not find database.", "OK", NULL);
    return;
  }

  if ((email = getWidgetString(addressInput)) == NULL || *email == '\0') {
    dialogString("Need an email address.", "OK", NULL);
    return;
  }
  lowerAddress(email);

  ptr = getWidgetString(majorInput);
  if (*ptr == '\0')
    Major = NULL;
  else
    Major = ptr;

  ptr = getWidgetString(minorInput);
  if (*ptr == '\0')
    Minor = NULL;
  else
    Minor = ptr;

  /*
   * Check if this is in the database.
   */
  if (!checkAddress(dbase, email)) {
    dialogString("This name is not in the database.", "Ok", NULL);
    return;
  }

  if (userData == NULL)
    dialogString("Are you sure about this?", "Delete", "No");
  else
    deleteEntry(dbase, email, Major, Minor);
}

/*************************************************************************
 *
 *************************************************************************/
#if NeedFunctionPrototypes
void pushQuit(Widget button, XtPointer userData, XtPointer call_data)
#else
void pushQuit(button, userData, call_data)
  Widget    button;
  XtPointer userData, call_data;
#endif
{
  dialogString("Are you sure?", "Quit", "Continue");
}

/*************************************************************************
 *
 *************************************************************************/
#if NeedFunctionPrototypes
void dialogButton(Widget button, XtPointer userData, XtPointer call_data)
#else
void dialogButton(button, userData, call_data)
  Widget    button;
  XtPointer userData, call_data;
#endif
{
  XtPopdown(dialog);
  XtDestroyWidget(dialog);

  /* Quit the application */
  if (strcmp(userData, "Quit") == 0) {
    XtDestroyApplicationContext(app);
    exit(0);
  }

  /* Create the link */
  if (strcmp(userData, "Link") == 0)
    pushLink(button, userData, call_data);

  /* Delete the info */
  if (strcmp(userData, "Delete") == 0)
    pushDelete(button, userData, call_data);

  /* Add the info to the database, importing the image into the dbase. */
  if ((strcmp(userData, "Import") == 0) || (strcmp(userData, "Filename") == 0))
    pushAdd(button, userData, call_data);
}

/*************************************************************************
 *
 *************************************************************************/
#if NeedFunctionPrototypes
void dialogString(char *msg, char *but1, char *but2)
#else
void dialogString(msg, but1, but2)
  char   *msg, *but1, *but2;
#endif
{
  static Widget  box, button;
  Arg            args[2];

  if (useX) {
    dialog = XtCreatePopupShell("dialogString", topLevelShellWidgetClass,
				toplevel, NULL, 0);

    /* create box widget */
    XtSetArg(args[0], XtNlabel, (XtArgVal) msg);
    box = XtCreateManagedWidget("error", dialogWidgetClass, dialog, args, 1);
    
    /* Create the button. */
    XtSetArg(args[0], XtNlabel, (XtArgVal) but1);
    button = XtCreateManagedWidget("button", commandWidgetClass, box, args, 1);
    XtAddCallback(button, XtNcallback, dialogButton, but1);
    
    if (but2 != NULL) {
      XtSetArg(args[0], XtNlabel, (XtArgVal) but2);
      button = XtCreateManagedWidget("button", commandWidgetClass,
				     box, args, 1);
      XtAddCallback(button, XtNcallback, dialogButton, but2);
    }

    /* instantiate it now */
    XtRealizeWidget(dialog);
    XtPopup(dialog, XtGrabExclusive);
  } else
    if (!noTxt)
      fprintf(stderr, "%s: %s\n", progname, msg);
}

/*************************************************************************
 *
 *************************************************************************/
#if NeedFunctionPrototypes
char *getWidgetString(Widget widget)
#else
char *getWidgetString(widget)
  Widget widget;
#endif
{
  Arg   args[1];
  char *s;
  
  XtSetArg(args[0], XtNstring, &s);
  XtGetValues(widget, args, 1);
  return(s);
}

/*************************************************************************
 *
 *************************************************************************/
#if NeedFunctionPrototypes
void setWidgetString(Widget widget, char *s)
#else
void setWidgetString(widget, s)
  Widget  widget;
  char   *s;
#endif
{
  Arg   args[1];

  if ((widget != NULL) && (s != NULL)) {
    XtSetArg(args[0], XtNstring, s);
    XtSetValues(widget, args, 1);
  }
}

/*************************************************************************
 *
 *************************************************************************/
#if NeedFunctionPrototypes
void usage(void)
#else
void usage()
#endif
{
  char *fmt;

  fmt = XtNewString("Usage: %s [-options] [-xoptions]\n\
       -v                   : Print the version number.\n\
       -list                : List the image(s) in the database.\n\
       -add                 : Ann an image to the database.\n\
       -delete              : Delete an image from the database.\n\
       -extract             : Extracts information from the database and\n\
                              writes it to disk.\n\
       -link                : Links an image in the database.\n\
       -dbase   <filename>  : The name of the database to use.\n\
       -email   <address>   : The name of the person.\n\
       -image   <filename>  : The filename you want to use.\n\
       -major   <type>      : These two form the MIME type, as in\n\
       -minor   <type>      :  image/gif, audio/sun.\n\
       -convert <directory> : Name of old style directory to convert.\n\
       -linkto  <address>   : Name of person of which image to get.\n\n\
  Besides these options the regular X options are recognized.\n");

  fprintf(stderr, fmt, progname);

  XtFree(fmt);
  exit(1);
}

